module Plots exposing (..)

-- Html
import Html exposing (Html)
import Html.Attributes as HA
import Html.Events exposing (onClick, onInput)
-- For Interactions
import Browser
-- Scatterplot stuff
import Axis
import Scale exposing (ContinuousScale)
import Statistics
import TypedSvg exposing (circle, g, rect, style, svg, text_)
import TypedSvg.Attributes exposing (class, fontFamily, fontSize, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, height, r, width, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Transform(..))
import TypedSvg.Events
import Color
--own
import Util exposing (..)

-- MODEL

type alias Model =
  { scatterplotType : ScatterplotType
  , plotChange : PlotChangeModel
  }


init : Model
init = { scatterplotType = None ""
       , plotChange = initPlotChange
       }

-- UPDATE

type Msg
  = ScatterplotChange ScatterplotType
  | ParallelCoordsSwapD_list Int Int
  | PlotChange PlotChangeMsg

noUpdate : Model -> Model
noUpdate model = model

update : Msg -> Model -> Model
update msg model =
  case msg of
    ScatterplotChange spType     -> { model | scatterplotType = spType  }
    ParallelCoordsSwapD_list i1 i2 ->
      { model | scatterplotType =
                  case model.scatterplotType of
                    None _                   -> None "Exception: triggered CarType2Change with PlotType None"
                    QQ_Plot                  -> None "Exception: triggered ParallelCoordsSwapD_list with PlotType QQ_Plot"
                    Parallel_Coords b d_list -> Parallel_Coords b (list_swap d_list {i1=i1,i2=i2})
                    ColoredShapes            -> None "Exception: triggered ParallelCoordsSwapD_list with PlotType ColoredShapes"
              , plotChange = let plotChange = model.plotChange
                             in
                             if      List.member i1 model.plotChange.selectedDims
                             && not (List.member i2 model.plotChange.selectedDims)
                             then { plotChange | selectedDims = List.reverse (updatePlotChange (SelectDim i2) <| updatePlotChange (SelectDim i1) model.plotChange).selectedDims }
                             else if      List.member i2 model.plotChange.selectedDims
                                  && not (List.member i1 model.plotChange.selectedDims)
                             then { plotChange | selectedDims = List.reverse (updatePlotChange (SelectDim i1) <| updatePlotChange (SelectDim i2) model.plotChange).selectedDims }
                             else if List.member i1 model.plotChange.selectedDims
                                  && List.member i2 model.plotChange.selectedDims
                             then { plotChange | selectedDims = List.reverse plotChange.selectedDims }
                             else model.plotChange
      }
    PlotChange plotChangeMsg -> { model | plotChange = updatePlotChange plotChangeMsg model.plotChange }

-- update plotChange

type alias PlotChangeModel = { selectedDims : List Int }

initPlotChange : PlotChangeModel
initPlotChange = { selectedDims = [] }

type PlotChangeMsg
  = SelectDim Int
  | ChangeToQQPlot Int

updatePlotChange : PlotChangeMsg -> PlotChangeModel -> PlotChangeModel
updatePlotChange msg model =
  case msg of
    SelectDim dim -> { model | selectedDims =
                                 case model.selectedDims of
                                   []    -> [dim]
                                   [x]   -> if dim == x
                                            then []
                                            else [x,dim]
                                   [x,y] -> if dim == x
                                            then [y]
                                            else if dim == y
                                                 then [x]
                                                 else [x,y]
                                   x::(y::xs) -> [x,y]
                     }
    ChangeToQQPlot dim -> model
-- Types
type ScatterplotType
    = None String
    | QQ_Plot
    | Parallel_Coords Bool (List XyData)
    | ColoredShapes

type alias Point =
    { pointName : String, x : Float, y : Float }


type alias XyData =
    { xDescription : String
    , yDescription : String
    , data : List Point
    }

type alias MultiDimPoint =
    { pointName : String, value : List Float }


type alias MultiDimData =
    { dimDescription : List String
    , data : List MultiDimPoint
    }

type alias MultiDimNominalPoint =
    { pointName : String, value : List String }

type alias MultiDimNominalData =
    { dimValDescription : List (List String) -- Description per dimension per nominal Value
    , dimDescriptions : List String          -- Description of the dimensions
    , data : List (List MultiDimNominalPoint) -- grouped points
    }

type alias ColoredShape =
    { colors : List (List Color.Color)   -- Color per field per nominal Value
    , fields : List (List (Float,Float)) -- list of polygon coordinates. Every list builds a field. All fields together build the shape.
    -- the total width and hight of the shape
    , w : Float
    , h : Float
    }

-- Eigenschaften

w : Float
w = 900


h : Float
h = 450


padding : Float
padding = 60


radius : Float
radius = 5.0

tickCount : Int
tickCount = 6

-- VIEW

view : Model -> Html Msg
view model =
    let bla = 0

    in Html.div []
            [
            ]

viewQQPlot : String -> XyData -> Html msg
viewQQPlot scatterplotName plotModel =
    viewScatterplot scatterplotName plotModel
      Nothing Nothing Nothing (Just "x=y") ""--(Just <| encode2Points <| getFirstThirdQuantilePoints plotModel) ""

viewScatterplot : String -> XyData -> Maybe XyData -> Maybe XyData -> Maybe XyData -> Maybe String -> String -> Html msg
viewScatterplot name model lowModel highModel otherModel line styleKind =
    Html.div []
      [ Html.h4 [] [Html.text name]
      , scatterplot model lowModel highModel otherModel line styleKind
      ]

viewParallelCoordsPlot : String -> MultiDimData -> Bool -> Bool -> List Int -> Html Msg
viewParallelCoordsPlot name model doXRay doSameScale selectedDims =
    Html.div []
      [ Html.h4 [] [Html.text name]
      , if List.length model.data /= 0 then parallelCoordsPlot model doXRay doSameScale selectedDims
        else Html.text "Unable to show the parallel coords plot, because the input data is empty!"
      ]

viewColoredShapesPlot : String -> MultiDimNominalData -> ColoredShape -> Html Msg
viewColoredShapesPlot name model shapeDefinition =
    Html.div []
      [ Html.h4 [] [Html.text name]
      , if List.length model.data /= 0                    -- must have at least 1 group
        then if List.length (list_get model.data 0 ) /= 0 -- the 1st group must have at least 1 point
        then
          Html.div []
            [ Html.p [] [Html.text "Legende:"]
            , Html.div []
                ( List.map3 (\dimName dimValues dimColors ->
                    Html.div []
                      [ Html.text <| dimName++": "++
                        (List.foldl (++) "" <| List.map2 (\value color ->
                                                 value++": "++(colorToString color)++", "
                                               ) dimValues dimColors
                        )
                      ]
                  ) model.dimDescriptions model.dimValDescription shapeDefinition.colors
                )
            , coloredShapesPlot model shapeDefinition
            ]

        else Html.text "Unable to show the colored shapes plot, because the input data is empty!"
        else Html.text "Unable to show the colored shapes plot, because the input data is empty!"
      ]

-- Plots
coloredShapesPlot : MultiDimNominalData -> ColoredShape -> Svg Msg
coloredShapesPlot model shapeDefinition =
  let thisw = (shapeDefinition.w+1) * (toFloat <| Maybe.withDefault 0 <| List.maximum <| List.map List.length model.data {-size of biggest group-})
      thish = (shapeDefinition.h+1) * (toFloat <| List.length model.data {-number of groups-})
  in
  svg [ viewBox 0 0 (thisw+2.5*padding) (thish+2*padding)
      , TypedSvg.Attributes.width <| TypedSvg.Types.Px (thisw+2.5*padding), TypedSvg.Attributes.height <| TypedSvg.Types.Px (thish+2*padding)
      ] [ TypedSvg.rect [ x 0
                        , y 0
                        , width thisw
                        , height thish
                        , TypedSvg.Attributes.fill   (TypedSvg.Types.Paint Color.white)
                        , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint Color.lightGrey)
                        , transform [ Translate (padding/2) padding ]
                        ] []
      , style [] [ TypedSvg.Core.text """
             .BigHoverDescription g + text { display: none; }
             .BigHoverDescription:hover g + text { display: inline; }
             .HoverDescription text { display: none; }
             .HoverDescription:hover text { display: inline; }
             .Button:hover { fill: rgb(0, 0, 0) }
             """]
      , let drawShape : MultiDimNominalPoint -> List (List String) -> List (List Color.Color) -> List (List (Float,Float)) -> (Float,Float) -> Svg Msg
            drawShape p dimValDescriptions dimValColors relShapeCoords (xShift,yShift)=
              g [ transform [ Translate (0.5*padding+xShift) (padding+yShift) ], class ["BigHoverDescription"] ]
                (List.append
                  [ g [] []
                  , text_ [ textAnchor TypedSvg.Types.AnchorStart
                                , fontFamily [ "sans-serif" ]
                                , fontSize <| TypedSvg.Types.Px 10
                                , x (thisw-xShift)
                                , y 10
                                ] [ TypedSvg.Core.text p.pointName ]
                  ]
                  ( List.map4 (\dimValue dimDescriptions dimColors dimCoords ->
                      let valueIndex = list_getIndexOf dimValue dimDescriptions
                      in
                      g [ class ["HoverDescription"] ]
                        [ TypedSvg.polygon
                          [ TypedSvg.Attributes.points dimCoords -- that are only rel coords!! WIP
                          , TypedSvg.Events.onClick (PlotChange <| ChangeToQQPlot <| list_getIndexOf dimValue p.value {-index of dim-})
                          , TypedSvg.Attributes.fill   (TypedSvg.Types.Paint <| list_get dimColors valueIndex)
                          , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint Color.black)
                          , TypedSvg.Attributes.strokeOpacity (TypedSvg.Types.Opacity 0.5)
                          , class [ "Button" ]
                          ] []
                        , g []
                            [ text_ [ textAnchor TypedSvg.Types.AnchorStart
                                    , fontFamily [ "sans-serif" ]
                                    , fontSize <| TypedSvg.Types.Px 10
                                    , x (thisw-xShift)
                                    , y 20
                                    ] [ TypedSvg.Core.text dimValue ]
                            ]
                        ]
                    ) p.value dimValDescriptions dimValColors relShapeCoords
                  )
                )

        in g []
             [ g []
                 ( List.map2 (\group yShift ->
                     g []
                       ( List.map2 (\p xShift ->
                           drawShape p model.dimValDescription shapeDefinition.colors shapeDefinition.fields (xShift,yShift)
                         ) group (List.map (toFloat >> (*) (shapeDefinition.w + 1)) <| List.range 0 <| List.length group)
                       )
                    ) model.data (List.map (toFloat >> (*) (shapeDefinition.h + 1)) <| List.range 0 <| List.length model.data)
                 )
             , drawShape (MultiDimNominalPoint "Legende" model.dimDescriptions)
                 (List.map (\a -> [a]) model.dimDescriptions)
                 (List.repeat (List.length shapeDefinition.fields) [Color.white])
                 shapeDefinition.fields (0,-padding)
             ]



      ]
parallelCoordsPlot : MultiDimData -> Bool -> Bool -> List Int -> Svg Msg
parallelCoordsPlot model doXRay doSameScale selectedDims =
    let listDimValues : List (List Float)
        listDimValues = multiDimDataToListDimValues model.data

        listDimScaleLocal : List (ContinuousScale Float)
        listDimScaleLocal = List.map (\dimValues -> yScale dimValues) listDimValues

        numberDims : Int
        numberDims = List.length listDimValues

        xCoordOfDim : Int -> Float
        xCoordOfDim index = (toFloat index)*(w/(toFloat numberDims))

        frontgroundColor = if doXRay then Color.white else Color.black
        backgroundColor = if doXRay then Color.black else Color.white

        sameScaleValues : List Float
        sameScaleValues = List.concat listDimValues

        sameScale : ContinuousScale Float
        sameScale = yScale sameScaleValues
    in
    svg [ viewBox 0 0 (w*1.1) (h*1.1), TypedSvg.Attributes.width <| TypedSvg.Types.Percent 90, TypedSvg.Attributes.height <| TypedSvg.Types.Percent 90 ]
        (List.concat
        [[ style [] [ TypedSvg.Core.text """
             .HoverDescription g > text { display: none; }
             .HoverDescription:hover g > text { display: inline; }
             .Button:hover { fill-opacity: 0.2 }
             """]
         ]
        , [ TypedSvg.rect [ x 0
                          , y 0
                          , width w
                          , height h
                          , TypedSvg.Attributes.fill (TypedSvg.Types.Paint backgroundColor)
                          ] [] ]
        , List.map3
            (\dimValues dimDescription index ->
                let usedColor = if List.member index selectedDims then Color.green else frontgroundColor
                in
                g [ transform [ Translate (padding - 1 + (xCoordOfDim index)) padding ]
                  , class [ "HoverDescription" ]
                  , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint usedColor)
                  , TypedSvg.Attributes.fill   (TypedSvg.Types.Paint usedColor)
                  , TypedSvg.Attributes.strokeWidth (Px 0.5)
                  , TypedSvg.Events.onClick (PlotChange <| SelectDim index)
                  ] [ yAxis <| if doSameScale then sameScaleValues else  dimValues
                    , text_ [ fontFamily [ "sans-serif" ]
                            , fontSize <| TypedSvg.Types.Px 10
                            , x -(padding/2)
                            , y ((-(padding/5))-((toFloat <| modBy 2 index)*padding/5))
                            ] [ TypedSvg.Core.text dimDescription ]
                    , if List.member index selectedDims then
                      text_ [ fontFamily [ "sans-serif" ]
                            , fontSize <| TypedSvg.Types.Px 10
                            , x -(padding/2)
                            , y (h-9/5*padding)
                            ] [ TypedSvg.Core.text <| "Selection "++(if index == (Maybe.withDefault -1 <|List.head selectedDims)
                                                                     then "1" else "2"
                                                                    )
                              ]
                      else text_ [] []
                    ]
            ) listDimValues model.dimDescription (List.range 0 <| numberDims-1)
        , let triangle_left: {x:Float,y:Float} -> Int -> Svg Msg
              triangle_left midPoint currentIndex =
                TypedSvg.polygon
                  [ TypedSvg.Attributes.points [(midPoint.x-2*radius,midPoint.y)
                                               ,(midPoint.x+(sqrt 3.0/4.0)*2*radius,midPoint.y+2*radius/2)
                                               ,(midPoint.x+(sqrt 3.0/4.0)*2*radius,midPoint.y-2*radius/2)
                                               ]
                  , TypedSvg.Events.onClick (ParallelCoordsSwapD_list (currentIndex-1) currentIndex)
                  , TypedSvg.Attributes.fill   (TypedSvg.Types.Paint frontgroundColor)
                  , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint frontgroundColor)
                  , class [ "Button" ]
                  ] []
              triangle_right: {x:Float,y:Float} -> Int -> Svg Msg
              triangle_right midPoint currentIndex =
                TypedSvg.polygon
                  [ TypedSvg.Attributes.points [(midPoint.x+2*radius,midPoint.y)
                                               ,(midPoint.x-(sqrt 3.0/4.0)*2*radius,midPoint.y+2*radius/2)
                                               ,(midPoint.x-(sqrt 3.0/4.0)*2*radius,midPoint.y-2*radius/2)
                                               ]
                  , TypedSvg.Events.onClick (ParallelCoordsSwapD_list currentIndex (currentIndex+1))
                  , TypedSvg.Attributes.fill   (TypedSvg.Types.Paint frontgroundColor)
                  , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint frontgroundColor)
                  , class [ "Button" ]
                  ] []
          in
          List.map
            (\dimIndex ->
                let usedColor = if List.member dimIndex selectedDims then Color.green else frontgroundColor
                in
                g [ transform [ Translate padding padding ]
                  ] ( List.append
                      ( if not (dimIndex > 0) then [] else
                        [ triangle_left  {x = (xCoordOfDim dimIndex)-2*radius,y = 0} dimIndex]
                      )(if not (dimIndex < numberDims-1) then [] else
                        [ triangle_right {x = (xCoordOfDim dimIndex)+2*radius,y = 0} dimIndex]
                      )
                    )
            ) (List.range 0 <| numberDims-1)
        , List.map
            (\point_ ->
                g [ transform [ Translate padding padding ] ]
                  [ TypedSvg.polyline
                    [ TypedSvg.Attributes.points
                      (List.map3
                        (\pDimValue dimScaleLocal index ->
                             ( xCoordOfDim index
                             , Scale.convert (if doSameScale then sameScale else dimScaleLocal) pDimValue
                             )
                        )
                        point_ listDimScaleLocal (List.range 0 <| numberDims-1)
                      )
                    , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint frontgroundColor)
                    , TypedSvg.Attributes.strokeWidth <| if doXRay then Px 2 else Px 0.3
                    , TypedSvg.Attributes.strokeOpacity (TypedSvg.Types.Opacity <| if doXRay then 0.1 else 1)
                    , TypedSvg.Attributes.fill TypedSvg.Types.PaintNone
                    ] []
                  ]
            ) <| List.map .value model.data
        ])

scatterplot : XyData -> Maybe XyData -> Maybe XyData -> Maybe XyData -> Maybe String -> String -> Svg msg
scatterplot model lowModel highModel otherModel line styleKind =
    let
        xValues : List Float
        xValues = List.map .x model.data

        yValues : List Float
        yValues = List.map .y model.data

        xScaleLocal : ContinuousScale Float
        xScaleLocal = xScale xValues

        yScaleLocal : ContinuousScale Float
        yScaleLocal = yScale yValues

        half : ( Float, Float ) -> Float
        half t = (Tuple.second t - Tuple.first t) / 2
    in
    svg [ viewBox 0 0 w h, TypedSvg.Attributes.width <| TypedSvg.Types.Percent 90, TypedSvg.Attributes.height <| TypedSvg.Types.Percent 90 ]
        (List.concat
        [[ style [] [ TypedSvg.Core.text ("""
             .point circle { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .point text { display: none; }
             .point:hover circle { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .point:hover text { display: inline; }
             """++{-Color-----------------------------------------------------------------}"""
             .lowC circle { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .lowC text { display: none; }
             .lowC:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .lowC:hover text { display: inline; }

             .highC circle { stroke: rgba(255, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .highC text { display: none; }
             .highC:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .highC:hover text { display: inline; }

             .otherC circle { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .otherC text { display: none; }
             .otherC:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .otherC:hover text { display: inline; }
             """++{-Size------------------------------------------------------------------}"""
             .lowS circle { stroke: rgba(0, 0, 0, 1); fill: rgba(255, 255, 255,1); stroke-width: 0.3;}
             .lowS text { display: none; }
             .lowS:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .lowS:hover text { display: inline; }

             .highS circle { stroke: rgba(0, 0, 0, 1); fill: rgba(255, 255, 255,1); stroke-width: 10; }
             .highS text { display: none; }
             .highS:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .highS:hover text { display: inline; }

             .otherS circle { stroke: rgba(0, 0, 0, 1); fill: rgba(255, 255, 255,1); stroke-width: 0.3;}
             .otherS text { display: none; }
             .otherS:hover circle { stroke: rgba(255, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .otherS:hover text { display: inline; }
             """++{-Form------------------------------------------------------------------}"""
             .lowF circle { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .lowF text { display: none; }
             .lowF:hover circle { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .lowF:hover text { display: inline; }

             .highF polygon { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .highF text { display: none; }
             .highF:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .highF:hover text { display: inline; }

             .otherF circle { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .otherF text { display: none; }
             .otherF:hover circle { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .otherF:hover text { display: inline; }
             """++{-Rotated-3angles-------------------------------------------------------}"""
             .lowR3 polygon { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .lowR3 text { display: none; }
             .lowR3:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .lowR3:hover text { display: inline; }

             .highR3 polygon { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .highR3 text { display: none; }
             .highR3:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .highR3:hover text { display: inline; }

             .otherR3 polygon { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .otherR3 text { display: none; }
             .otherR3:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .otherR3:hover text { display: inline; }
             """++{-Roteted-Lines---------------------------------------------------------}"""
             .lowRL polyline { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .lowRL text { display: none; }
             .lowRL:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .lowRL:hover text { display: inline; }

             .highRL polyline { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .highRL text { display: none; }
             .highRL:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .highRL:hover text { display: inline; }

             .otherRL polyline { stroke: rgba(0, 0, 0, 0.4); fill: rgba(255, 255, 255,0.3); }
             .otherRL text { display: none; }
             .otherRL:hover polygon { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
             .otherRL:hover text { display: inline; }
             """) ]
         , g [ transform [ Translate (padding - 1) (h - padding) ] ]
             [ xAxis (List.map (\onePoint -> onePoint.x) model.data)
             , text_ [ fontFamily [ "sans-serif" ]
                     , fontSize (TypedSvg.Types.Px 10)
                     , textAnchor TypedSvg.Types.AnchorMiddle
                     , x (w/2-padding)
                     , y (padding*2/3)
                     ] [ TypedSvg.Core.text model.xDescription ]
             ]
         , g [ transform [ Translate (padding - 1) padding ] ]
             [ yAxis (List.map (\onePoint -> onePoint.y) model.data)
             , text_ [ fontFamily [ "sans-serif" ]
                     , fontSize (TypedSvg.Types.Px 10)
                     , x -(padding/2)
                     , y -(padding/5)
                     --((Scale.convert yScaleLocal labelPositions.y)-padding*1/3)
                     ] [ TypedSvg.Core.text model.yDescription ]
              ]
         ]
        , if ((maybeIsNothing lowModel) || (maybeIsNothing highModel) || (maybeIsNothing otherModel))
          then [ g [ transform [ Translate padding padding ] ]
                   (List.map (point "point" xScaleLocal yScaleLocal) model.data)
               ]
          else let lowP   = (Maybe.withDefault (XyData "" "" [])   lowModel).data
                   highP  = (Maybe.withDefault (XyData "" "" [])  highModel).data
                   otherP = (Maybe.withDefault (XyData "" "" []) otherModel).data
               in
                 [g [ transform [ Translate padding padding ] ]
                     (List.map (point ("low"++styleKind) xScaleLocal yScaleLocal) lowP)
                 , g [ transform [ Translate padding padding ] ]
                     (List.map (point ("high"++styleKind) xScaleLocal yScaleLocal) highP)
                 , g [ transform [ Translate padding padding ] ]
                     (List.map (point ("other"++styleKind) xScaleLocal yScaleLocal) otherP)
                 ]
        , case line of
            Just lineOption ->
              case lineOption of
                "x=y" -> -- build a line x=y
                    let extendX = wideExtent xValues
                        extendY = wideExtent yValues
                        maxValue = min (Tuple.second extendX) (Tuple.second extendY)
                        minValue = max (Tuple.first  extendX) (Tuple.first  extendY)
                    in
                    [ g [ transform [ Translate padding padding ] ]
                       [ TypedSvg.polyline
                          [ TypedSvg.Attributes.points
                              [ (Scale.convert xScaleLocal minValue, Scale.convert yScaleLocal minValue)
                              , (Scale.convert xScaleLocal maxValue, Scale.convert yScaleLocal maxValue)
                              ]
                          , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint Color.black)
                          , TypedSvg.Attributes.strokeOpacity (TypedSvg.Types.Opacity 1)
                          ] []
                       ]
                    ]
                _     ->
                  case (String.toFloat lineOption) of
                    Just floatValue ->
                     -- build a line x = avgX
                     let avgX = Maybe.withDefault 0 (String.toFloat lineOption)
                     in
                     [ g [ transform [ Translate padding padding ] ]
                         [ TypedSvg.polyline
                            [ TypedSvg.Attributes.points
                                [ ((Scale.convert xScaleLocal avgX)-radius, 0)
                                , ((Scale.convert xScaleLocal avgX)-radius, h-2*padding)
                                ]
                            , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint Color.black)
                            , TypedSvg.Attributes.strokeOpacity (TypedSvg.Types.Opacity 0.3)
                            ] []
                         ]
                     ]
                    Nothing         ->
                     -- try to parse 2 coords for the line -> expect: x1,y1|x2,y2
                     let maytwoCoords = decode2Points lineOption
                     in
                     case maytwoCoords of
                       Nothing        -> []
                       Just twoCoords ->
                         let minMaxLineCoords = calculateMinMaxLineCoords
                                                    (twoCoords.p1,twoCoords.p2)
                                                    (wideExtent xValues)
                                                    (wideExtent yValues)
                             newx1 = (Tuple.first minMaxLineCoords ).x
                             newy1 = (Tuple.first minMaxLineCoords ).y
                             newx2 = (Tuple.second minMaxLineCoords).x
                             newy2 = (Tuple.second minMaxLineCoords).y
                         in
                         [ g [ transform [ Translate padding padding ] ]
                             [ TypedSvg.polyline
                                [ TypedSvg.Attributes.points
                                    [ ((Scale.convert xScaleLocal newx1), (Scale.convert yScaleLocal newy1))
                                    , ((Scale.convert xScaleLocal newx2), (Scale.convert yScaleLocal newy2))
                                    ]
                                , TypedSvg.Attributes.stroke (TypedSvg.Types.Paint Color.black)
                                , TypedSvg.Attributes.strokeOpacity (TypedSvg.Types.Opacity 1)
                                ] []
                             ]
                         ]
            Nothing -> []
        ])

-- Axis and Scale

xAxis : List Float -> Svg msg
xAxis values =
    Axis.bottom [ Axis.tickCount tickCount ] (xScale values)


yAxis : List Float -> Svg msg
yAxis values =
    Axis.left [ Axis.tickCount tickCount ] (yScale values)

xScale : List Float -> ContinuousScale Float
xScale values =
    Scale.linear ( 0, w - 2 * padding ) ( wideExtent values )


yScale : List Float -> ContinuousScale Float
yScale values =
    Scale.linear ( h - 2 * padding, 0 ) ( wideExtent values )

wideExtent : List Float -> ( Float, Float )
wideExtent values =
    let datenbereich = Maybe.withDefault (0,0) (Statistics.extent values)
        zugabe = ((Tuple.second datenbereich)-(Tuple.first datenbereich))/(toFloat (2*tickCount))
    in (max 0 (Tuple.first datenbereich-zugabe), (Tuple.second datenbereich+zugabe))

-- Plot-values

point : String -> ContinuousScale Float -> ContinuousScale Float -> Point -> Svg msg
point pointClass scaleX scaleY xyPoint =
  let
    kreisbeschriftung : String
    kreisbeschriftung = xyPoint.pointName++
      (" ("++((String.fromFloat xyPoint.x)++(","++((String.fromFloat xyPoint.y)++")"))))
    xValue = Scale.convert scaleX xyPoint.x
    yValue = Scale.convert scaleY xyPoint.y
  in
    g [ class [ pointClass ], fontSize <| Px 15.0, fontFamily [ "sans-serif" ] ]
        [ if (pointClass == "highF" || pointClass == "highR3") then
              -- 3angle mit Spitze nach oben
              TypedSvg.polygon
                  [ TypedSvg.Attributes.points
                      [ (xValue,yValue-radius)
                      , (xValue-(radius/(sqrt 2)),yValue+(radius/(sqrt 2)))
                      , (xValue+(radius/(sqrt 2)),yValue+(radius/(sqrt 2)))
                      ]
                  ] []
          else if (pointClass == "lowR3" || pointClass == "otherR3") then
              -- 3angle mit Spitze nach unten
              TypedSvg.polygon
                  [ TypedSvg.Attributes.points
                      [ (xValue,yValue+radius)
                      , (xValue-(radius/(sqrt 2)),yValue-(radius/(sqrt 2)))
                      , (xValue+(radius/(sqrt 2)),yValue-(radius/(sqrt 2)))
                      ]
                  ] []
          else if (pointClass == "highRL") then
              -- Linie vertikal
              TypedSvg.polyline
                  [ TypedSvg.Attributes.points
                      [ (xValue,yValue-radius)
                      , (xValue,yValue+radius)
                      ]
                  ] []
          else if (pointClass == "lowRL" || pointClass == "otherRL") then
              -- Linie horizontal
              TypedSvg.polyline
                  [ TypedSvg.Attributes.points
                      [ (xValue-radius,yValue)
                      , (xValue+radius,yValue)
                      ]
                  ] []
          else circle [ cx xValue
                      , cy yValue
                      , r radius  ][]
        , text_ [ textAnchor TypedSvg.Types.AnchorMiddle
                , x xValue
                , y (yValue-15)
                ] [TypedSvg.Core.text kreisbeschriftung]
        ]
-- Math
calculateMinMaxLineCoords : (Point,Point) -> (Float,Float) -> (Float,Float) -> (Point,Point)
calculateMinMaxLineCoords input extendX extendY =
  let x1 = (Tuple.first input).x
      y1 = (Tuple.first input).y
      x2 = (Tuple.second input).x
      y2 = (Tuple.second input).y
      -- line: y = m*x + n
      m = (y2-y1)/(x2-x1)
      n = y1-m*x1
      -- calculate the case of what scatterplot edges the line touches
      minX = Tuple.first extendX
      maxX = Tuple.second extendX
      minY = Tuple.first extendY
      maxY = Tuple.second extendY
      minCase = if m*minX+n > minY then "y"{-touches left -} else "x"{-touches lower-}
      maxCase = if m*maxX+n > maxY then "x"{-touches upper-} else "y"{-touches right-}
      -- calculate new coords depending with respect to the case
      newx1 = if minCase == "x" then (minY-n)/m else minX
      newy1 = m*newx1+n
      newx2 = if maxCase == "x" then (maxY-n)/m else maxX
      newy2 = m*newx2+n
  in (Point "" newx1 newy1, Point "" newx2 newy2)

powerTransformation : Float -> List Point -> List Point
powerTransformation power oldPs =
  if power == 0
  then List.map (\p -> Point p.pointName (logBase 2 p.x) (logBase 2 p.y)) oldPs
  else List.map (\p -> Point p.pointName (p.x^power    ) (p.y^power    )) oldPs

getFirstThirdQuantilePoints : XyData -> {p1: Point,p2: Point}
getFirstThirdQuantilePoints model =
  let dividedCoords = pointsDeConstructor model.data
  in{p1 = Point "first-Q" (Maybe.withDefault -1 (Statistics.quantile 0.25 dividedCoords.xs))
                          (Maybe.withDefault -1 (Statistics.quantile 0.25 dividedCoords.ys))
    ,p2 = Point "third-Q" (Maybe.withDefault -1 (Statistics.quantile 0.75 dividedCoords.xs))
                          (Maybe.withDefault -1 (Statistics.quantile 0.75 dividedCoords.ys))
    }

{- Requires both inputs to be Q-Values
 -}
qValuesToQQValues : XyData -> XyData -> XyData
qValuesToQQValues qValues1 qValues2 =
    let isSmallerQValues1 = (List.length qValues1.data) < (List.length qValues2.data)
        -- Unterscheidung in smaller und greater für optimale Auflösungsanpassung
        smaller = if isSmallerQValues1 then qValues1 else qValues2
        greater = if isSmallerQValues1 then qValues2 else qValues1
        -- Umwandlung in verwendbare Form
        qV1 = List.map (\aPoint -> aPoint.y) qValues1.data
        qV2 = List.map (\aPoint -> aPoint.y) qValues2.data

    in XyData qValues2.yDescription qValues1.yDescription
         <| List.filterMap
              (\sPoint ->
                   Maybe.map3
                           pointConstructor
                           (Just "")
                           (Statistics.quantile sPoint.x qV2)
                           (Statistics.quantile sPoint.x qV1)
              ) smaller.data

qValuesToMultiDimQQValues : List XyData -> MultiDimData
qValuesToMultiDimQQValues qList =
    let smallestDimData : XyData
        smallestDimData = list_get (List.sortBy (\xyData -> List.length <| xyData.data) qList) 0
        -- gets all values and description for one dimension
        getOneDimData : XyData -> {description : String, value : List Float}
        getOneDimData dimension_data =
            { description = dimension_data.yDescription
            , value       =
                List.map
                  (\quantil -> Maybe.withDefault 999 <| Statistics.quantile quantil <| List.map .y dimension_data.data)
                  (List.map .x smallestDimData.data)
            }
        -- gets all the data that is nessesary
        multiDimValues : List {description : String, value : List Float}
        multiDimValues = List.map (\xyData -> getOneDimData xyData) qList
    in MultiDimData (List.map .description multiDimValues) <| listDimValuesToMultiDimData <| List.map .value multiDimValues

listDimValuesToMultiDimData : List (List Float) -> List MultiDimPoint
listDimValuesToMultiDimData ldv =
    let multiDimPointsConstructor : List (List Float) -> List (List Float)-> List MultiDimPoint
        multiDimPointsConstructor dimList pointList =
          case dimList of
            []        -> List.map (\p -> MultiDimPoint "" p) pointList
            dim::dims -> multiDimPointsConstructor dims <| List.map2 (\dimValue p -> List.append p [dimValue]) dim pointList
    in  multiDimPointsConstructor ldv (List.repeat (List.length <| Maybe.withDefault [] <| List.head ldv) [])

multiDimDataToListDimValues : List MultiDimPoint -> List (List Float)
multiDimDataToListDimValues mdd =
    let mdv : List (List Float)
        mdv = List.map .value mdd

        listDimValuesConstructor : List (List Float) -> List (List Float) -> List (List Float)
        listDimValuesConstructor pointList dimList =
          case pointList of
            []             -> dimList
            point_::points -> listDimValuesConstructor points <| List.map2 (\pDimValue dim -> List.append dim [pDimValue]) point_ dimList
    in  listDimValuesConstructor mdv (List.repeat (List.length <| Maybe.withDefault [] <| List.head mdv) [])

encode2Points : {p1: Point, p2: Point} -> String -- Format: x1,y1|x2,y2
encode2Points ps = (String.fromFloat ps.p1.x)++","++(String.fromFloat ps.p1.y)++"|"++(String.fromFloat ps.p2.x)++","++(String.fromFloat ps.p2.y)

decode2Points : String -> Maybe {p1: Point, p2: Point} -- Format: x1,y1|x2,y2 ; expects: no coord on -1
decode2Points string =
  let coordsList : List String
      coordsList = String.split "|" string

      coord1List : List String
      coord1List = String.split "," (Maybe.withDefault "" (List.head coordsList))

      coord2List : List String
      coord2List = String.split "," (Maybe.withDefault "" (List.head (List.drop 1 coordsList)))

      x1 = Maybe.withDefault -1 (String.toFloat (Maybe.withDefault "" (List.head coord1List)))
      y1 = Maybe.withDefault -1 (String.toFloat (Maybe.withDefault "" (List.head (List.drop 1 coord1List))))

      x2 = Maybe.withDefault -1 (String.toFloat (Maybe.withDefault "" (List.head coord2List)))
      y2 = Maybe.withDefault -1 (String.toFloat (Maybe.withDefault "" (List.head (List.drop 1 coord2List))))
  in if (   List.length coordsList /= 2
         || List.length coord1List /= 2
         || List.length coord2List /= 2
         || x1 == -1 || y1 == -1 || x2 == -1 || y2 == -1
        )
     then Nothing
     else Just {p1 = Point "" x1 y1, p2 = Point "" x2 y2}

getDoXRayFromPlotType : ScatterplotType -> Bool
getDoXRayFromPlotType p_type = case p_type of
                                 None _                   -> False
                                 QQ_Plot                  -> False
                                 Parallel_Coords doXRay _ -> doXRay
                                 ColoredShapes            -> False

getD_listFromPlotType : ScatterplotType -> List XyData
getD_listFromPlotType p_type = case p_type of
                                 None _                   -> []
                                 QQ_Plot                  -> []
                                 Parallel_Coords _ d_list -> d_list
                                 ColoredShapes            -> []

getPlotChangeMsg : Msg -> Maybe PlotChangeMsg
getPlotChangeMsg msg =
  case msg of
    ScatterplotChange        _   -> Nothing
    ParallelCoordsSwapD_list _ _ -> Nothing
    PlotChange               m   -> Just m

msgToString : Msg -> String
msgToString msg =
  case msg of
    ScatterplotChange        _   -> "ScatterplotChange"
    ParallelCoordsSwapD_list _ _ -> "ParallelCoordsSwapD_list"
    PlotChange               _   -> "PlotChange"

plotChangeMsgToString : PlotChangeMsg -> String
plotChangeMsgToString msg =
  case msg of
    SelectDim      _ -> "SelectDim"
    ChangeToQQPlot _ -> "ChangeToQQPlot"

colorToString : Color.Color -> String
colorToString color =
       if color == Color.white then "white"
  else if color == Color.black then "black"
  else if color == Color.green then "green"
  else if color == Color.yellow then "yellow"
  else if color == Color.blue then "blue"
  else if color == Color.purple then "purple"
  else if color == Color.red then "red"
  else if color == Color.orange then "orange"
  else "this color is unknown for colorToString"

-- Constructors
pointConstructor : String -> Float -> Float -> Point
pointConstructor name x y = Point name x y

pointsConstructor : List Float -> List Float -> List Point
pointsConstructor xValues yValues =
    let onePointConstructor : List Float -> List Float -> List Point -> List Point
        onePointConstructor xVs yVs points=
          case xVs of
            []    ->
              case yVs of
                []    -> points
                y::ys -> []
            x::xs ->
              case yVs of
                []    -> []
                y::ys -> onePointConstructor xs ys
                           (List.append points [(Point "" x y)])
    in onePointConstructor xValues yValues []

multiDimNominalDataConstructor : List (List String) -> List String -> List (List (String, List String)) -> MultiDimNominalData
multiDimNominalDataConstructor dimValDescription dimDescriptions groups =
  MultiDimNominalData dimValDescription dimDescriptions
  (List.map (\group -> List.map (\(pointName,value) -> MultiDimNominalPoint pointName value
                       ) group
   ) groups
  )

coloredShapeConstructor : List (List Color.Color) -> List (List (Float,Float)) -> ColoredShape
coloredShapeConstructor colors fields =
  let maxW = Maybe.withDefault 0 <| List.maximum <| List.map (\field -> Maybe.withDefault 0 <| List.maximum <| List.map Tuple.first  field) fields
      minW = Maybe.withDefault 0 <| List.minimum <| List.map (\field -> Maybe.withDefault 0 <| List.minimum <| List.map Tuple.first  field) fields
      maxH = Maybe.withDefault 0 <| List.maximum <| List.map (\field -> Maybe.withDefault 0 <| List.maximum <| List.map Tuple.second field) fields
      minH = Maybe.withDefault 0 <| List.minimum <| List.map (\field -> Maybe.withDefault 0 <| List.minimum <| List.map Tuple.second field) fields
  in ColoredShape colors fields (maxW-minW) (maxH-minH)

-- DeConstructors
pointsDeConstructor : List Point -> {xs: List Float, ys: List Float}
pointsDeConstructor points =
  let pointsDeConstructorHelper : List Point -> {xs: List Float, ys: List Float} -> {xs: List Float, ys: List Float}
      pointsDeConstructorHelper pointsH r =
        case pointsH of
          []    -> r
          p::ps -> pointsDeConstructorHelper ps {r | xs = List.append [p.x] r.xs
                                                   , ys = List.append [p.y] r.ys
                                                }
  in pointsDeConstructorHelper points {xs = [], ys = []}

-- Working with maybe

maybeIsNothing : Maybe a -> Bool
maybeIsNothing maybe = case maybe of
                          Just sth -> False
                          Nothing  -> True

-- Working with Lists
listZip : ( List a, List b ) -> List ( a, b )
listZip tuple =
    List.map2 (\fst snd -> (fst,snd)) (Tuple.first tuple) (Tuple.second tuple)
