module StudentsPerformanceInExams exposing (main)

-- elm stuff
import Browser
import Html exposing (Html, text, pre, div, table, tr, th, td, button, hr)
import Html.Events exposing (onClick)
import Http
import Dict
import Maybe.Extra
-- Decode and Date, Time stuff
import Html.Parser
import Parser
import Csv
import Csv.Decode
import Date
import Time exposing (Month(..))
-- Svg stuff
import TypedSvg exposing (circle, g, rect, style, svg, text_)
import TypedSvg.Attributes exposing (class, fontFamily, fontSize, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, height, r, width, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Transform(..))
-- visualisation
import Scale exposing (ContinuousScale)
import Scale.Color
import Statistics
-- other
import Color
-- own
import Plots exposing (XyData, MultiDimData)
import Util exposing (..)

-- MAIN
main =
  Browser.element
    { init = init
    , update = update
    , subscriptions = subscriptions
    , view = view
    }

-- MODEL
type alias Model = { httpState : HttpState
                   , fullTexts : List {description: String, data: String}
                   , data : List Data
                   , error : String
                   , visualisationKind : VisualisationKind
                   , plotModel : Plots.Model
                   , doViewIntermediateResult : Bool
                   }

-- Types

type HttpState
  = Failure
  | Loading Int
  | Success

type Arrangement
  = Line_By_Line
  | Forth_Loop

--type alias OldData = {description: String, data: List ( String, Maybe Float )}

type alias Data = { gender: String
                  , race: String
                  , parentEducation: String
                  , lunch: String
                  , testPreparation: String
                  , mathScore: Int
                  , readingScore: Int
                  , writingScore: Int
                  , avgScore: Int
                  }

type alias DataDescription =
                  { gender: String
                  , race: String
                  , parentEducation: String
                  , lunch: String
                  , testPreparation: String
                  , mathScore: String
                  , readingScore: String
                  , writingScore: String
                  , avgScore: String
                  }

type VisualisationKind
 = Table String Bool
 | QQ_Plot                              (Data -> String) Int
 | MultiDimPlot (Maybe (XyData,XyData)) (Data -> String) Int
 | ColorIcons

filesToLoad =
  [ "StudentsPerformance.csv"
  ]

dataDescription : DataDescription
dataDescription = { gender          = "gender"
                  , race            = "race"
                  , parentEducation = "parentEducation"
                  , lunch           = "lunch"
                  , testPreparation = "testPreparation"
                  , mathScore       = "mathScore"
                  , readingScore    = "readingScore"
                  , writingScore    = "writingScore"
                  , avgScore        = "avgScore"
                  }

dataDescription2 : Data
dataDescription2 ={ gender          = "gender"
                  , race            = "race/ethnicity"
                  , parentEducation = "parental education"
                  , lunch           = "lunch"
                  , testPreparation = "test preparation"
                  , mathScore       = 0
                  , readingScore    = 0
                  , writingScore    = 0
                  , avgScore        = 0
                  }

dataPossibleNominalValues : List (List String)
dataPossibleNominalValues =[["male","female"]
                           ,["group A","group B","group C","group D","group E"]
                           ,["master's degree","bachelor's degree","associate's degree","some college","some high school","high school"]
                           ,["standard","free/reduced"]
                           ,["completed","none"]
                           ]

init : () -> (Model, Cmd Msg)
init _ =
  ( Model (Loading <| List.length filesToLoad) [] [] "" (Table "no" True) Plots.init False
  , Cmd.batch <|
      List.map (\file ->
                  Http.get
                    { url = file
                    , expect = Http.expectString <| GotText file
                    }
               ) filesToLoad
  )

-- UPDATE
type Msg
  = GotText String (Result Http.Error String)
  | ChangeVisualisation (VisualisationKind)
  | UpdatePlots (Plots.Msg)
  | ChangeDoViewIntermediateResult

update : Msg -> Model -> ( Model, Cmd Msg)
update msg model = (updateModel msg model, Cmd.none)

updateModel : Msg -> Model -> Model
updateModel msg model =
  case msg of
    GotText file result ->
      case result of
        Ok fullText ->
          case model.httpState of
            Failure           -> {model | httpState = Failure}
            Loading filesLeft ->
                if filesLeft-1 == 0 then
                    let newFullTexts = {description = file, data = fullText}::model.fullTexts
                        checkedFullTexts = List.map checkInputFormat newFullTexts
                    in
                    { model | httpState = Success
                            , fullTexts = checkedFullTexts
                            , data      = csvString_to_data <| Maybe.withDefault "" <| List.head <| List.map .data checkedFullTexts
                    }
                else{ model | httpState = Loading (filesLeft-1)
                            , fullTexts = {description = file, data = fullText}::model.fullTexts
                    }
            Success           -> {model | httpState = Failure}
        Err e -> { model | httpState = Failure
                         , error = httpErrorToString e
                 }
    ChangeVisualisation kind ->
      { model | visualisationKind = kind
              , plotModel = let initialPlot = Plots.init
                            in
                            { initialPlot | scatterplotType =
                                              case kind of
                                                Table   _ _                    -> Plots.None ""
                                                QQ_Plot _ _                    -> Plots.None ""
                                                ColorIcons                     -> Plots.None ""
                                                MultiDimPlot oldSelectDims s i -> Plots.Parallel_Coords False (calcQValues model s i)
                            }
      }
    UpdatePlots plotsMsg -> let updatedPlotsModel = Plots.update plotsMsg model.plotModel
                                plotChangeMsg = Plots.getPlotChangeMsg plotsMsg
                            in
                            case plotChangeMsg of
                              Just (Plots.SelectDim _) ->
                                case model.visualisationKind of
                                  Table   _ _                    -> model
                                  QQ_Plot _ _                    -> model
                                  ColorIcons                     -> model
                                  MultiDimPlot oldSelectDims s i ->
                                    if List.length updatedPlotsModel.plotChange.selectedDims == 2
                                    then let xyDataList = Plots.getD_listFromPlotType updatedPlotsModel.scatterplotType
                                             selectedDim1 = list_get xyDataList <| list_get updatedPlotsModel.plotChange.selectedDims 0
                                             selectedDim2 = list_get xyDataList <| list_get updatedPlotsModel.plotChange.selectedDims 1
                                         in
                                         { model | plotModel         = updatedPlotsModel
                                                 , visualisationKind = MultiDimPlot (Just (selectedDim1,selectedDim2)) s i
                                         }
                                    else { model | plotModel         = updatedPlotsModel
                                                 , visualisationKind = MultiDimPlot Nothing s i
                                         }
                              Just (Plots.ChangeToQQPlot dimIndex) ->
                                case dimIndex of
                                  0 -> updateModel (ChangeVisualisation (QQ_Plot              .gender          0)) model
                                  1 -> updateModel (ChangeVisualisation (MultiDimPlot Nothing .race            1)) model
                                  2 -> updateModel (ChangeVisualisation (MultiDimPlot Nothing .parentEducation 2)) model
                                  3 -> updateModel (ChangeVisualisation (QQ_Plot              .lunch           3)) model
                                  4 -> updateModel (ChangeVisualisation (QQ_Plot              .testPreparation 4)) model
                                  _ -> { model | httpState = Failure
                                               , error = ("ChangeToQQPlot failed with dimIndex = "++(String.fromInt dimIndex))
                                       }
                              Nothing -> { model | plotModel = updatedPlotsModel }
    ChangeDoViewIntermediateResult -> { model | doViewIntermediateResult = not model.doViewIntermediateResult }

-- SUBSCRIPTIONS
subscriptions : Model -> Sub Msg
subscriptions model = Sub.none

-- VIEW
view : Model -> Html Msg
view model =
  case model.httpState of
    Failure       -> text model.error
    Loading index -> text ("Loading... "++(String.fromInt index)++" left")
    Success ->
      div []
        [ viewHeader
        , case model.visualisationKind of
            Table sortBy asc -> viewFullDataTable model sortBy asc
            QQ_Plot selector dimIndex ->
              let dimName = selector dataDescription2
                  dimValues = list_get dataPossibleNominalValues dimIndex
                  dimValue0 = list_get dimValues 0
                  dimValue1 = list_get dimValues 1
                  dataSelection = selectData selector (list_get dimValues 0) model.data
                  pointsWithSelection    = dataToQValuesXyData dataSelection.selected   .avgScore .avgScore dimValue0
                  pointsWithoutSelection = dataToQValuesXyData dataSelection.unselected .avgScore .avgScore dimValue1
                  qqValues = Plots.qValuesToQQValues pointsWithSelection pointsWithoutSelection
              in
              div []
                [ div []
                    [ button [onClick ChangeDoViewIntermediateResult] [ text <| if model.doViewIntermediateResult then "Zwischenergebnisse ausblenden"
                                                                                                                  else "Zwischenergebnisse einblenden"
                                                                      ]
                    ]
                , let qqPlotName = ("QQPlot Students avgScore "++dimName++": "++dimValue0++" vs "++dimName++": "++dimValue1)
                  in Plots.viewQQPlot qqPlotName qqValues
                , if model.doViewIntermediateResult
                  then div []
                         [ Plots.viewScatterplot ("QPlot "++dimName++": "++dimValue0) pointsWithSelection    Nothing Nothing Nothing Nothing ""
                         , Plots.viewScatterplot ("QPlot "++dimName++": "++dimValue1) pointsWithoutSelection Nothing Nothing Nothing Nothing ""
                         ]
                  else div [] []
                ]
            MultiDimPlot may2D selector dimIndex ->
              let dimValues = list_get dataPossibleNominalValues dimIndex
                  qValues = calcQValues model selector dimIndex
                  -- Create MultiDimData
                  qqValues = Plots.qValuesToMultiDimQQValues <| Plots.getD_listFromPlotType model.plotModel.scatterplotType
              in
              div []
                [ div []
                    [ button [onClick ChangeDoViewIntermediateResult] [ text <| if model.doViewIntermediateResult then "Zwischenergebnisse ausblenden"
                                                                                                                  else "Zwischenergebnisse einblenden"
                                                                      ]
                    ]
                , div [] [text "Click on 2 axis to show them both in a 2D graph."]
                , viewInteractivePlot <| Plots.viewParallelCoordsPlot
                                           ("Parallel QQ-Plot for "++(selector dataDescription2)++" showing the avgScore")
                                           qqValues False True model.plotModel.plotChange.selectedDims
                , case may2D of
                    Just (xDim,yDim) ->
                      Plots.viewQQPlot ("QQ-Plot "++(xDim.yDescription)++" "++yDim.yDescription) (Plots.qValuesToQQValues yDim xDim)
                    Nothing -> div [] []
                , if model.doViewIntermediateResult
                  then div []
                         (List.map2 (\dimValue qValue ->
                            Plots.viewScatterplot ("QPlot with "++dimValue) qValue Nothing Nothing Nothing Nothing ""
                          ) dimValues qValues
                         )
                  else div [] []
                ]
            ColorIcons ->
              let groupedData : List (List Data)
                  groupedData = list_groupBy .avgScore model.data

                  data : Plots.MultiDimNominalData
                  data = Plots.multiDimNominalDataConstructor
                           dataPossibleNominalValues
                           ["gender"
                           ,"race/ethnicity"
                           ,"parental education"
                           ,"lunch"
                           ,"test preparation"
                           ] <| List.map (\group-> List.map (\line-> ("avg score "++(String.fromInt line.avgScore)
                                                                     , [line.gender
                                                                       ,line.race
                                                                       ,line.parentEducation
                                                                       ,line.lunch
                                                                       ,line.testPreparation
                                                                       ]
                                                                     )) group ) groupedData
                  shapeDefinition : Plots.ColoredShape
                  shapeDefinition = Plots.coloredShapeConstructor
                                      [[Color.blue,Color.red]
                                      ,[Color.yellow,Color.purple,Color.blue,Color.green,Color.white]
                                      ,[Color.red,Color.yellow,Color.purple,Color.blue,Color.green,Color.white]
                                      ,[Color.green,Color.white]
                                      ,[Color.green,Color.red]
                                      ]
                                        {- Shape test 1
                                        [[(0,0),(10,0),(10, 4),(0, 4)]
                                        ,[(0,4),( 5,4),( 5, 7),(0, 7)]
                                        ,[(5,4),(10,4),(10, 7),(5, 7)]
                                        ,[(0,7),( 5,7),( 5,10),(0,10)]
                                        ,[(5,7),(10,7),(10,10),(5,10)]
                                        ]-}
                                        [[(3,3),(8,3),(8,8),(3,8)]
                                        ,[(0  ,0  ),( 5.5,0  ),( 5.5,3  ),(  3, 3  ),(  3, 5.5),(0  , 5.5)]
                                        ,[(5.5,0  ),(11  ,0  ),(11  ,5.5),(  8, 5.5),(  8, 3  ),(5.5, 3  )]
                                        ,[(0  ,5.5),( 3  ,5.5),(   3,8  ),(5.5, 8  ),(5.5,11  ),(0  ,11  )]
                                        ,[(8  ,5.5),(11  ,5.5),(11  ,11 ),(5.5,11  ),(5.5, 8  ),(8  , 8  )]
                                        ]

              in viewInteractivePlot <| Plots.viewColoredShapesPlot
                                          "Überblick: Eigenschaften der Studenten nach Durchschnittsleistung sortiert"
                                          data shapeDefinition
        ]

-- big view Parts
viewFullDataTable : Model -> String -> Bool -> Html Msg
viewFullDataTable model sortBy asc =
  let sortedData1 = case sortBy of
                      "race"           -> List.sortBy .race            model.data
                      "parentEducation"-> List.sortBy .parentEducation model.data
                      "lunch"          -> List.sortBy .lunch           model.data
                      "testPreparation"-> List.sortBy .testPreparation model.data
                      "mathScore"      -> List.sortBy .mathScore       model.data
                      "readingScore"   -> List.sortBy .readingScore    model.data
                      "writingScore"   -> List.sortBy .writingScore    model.data
                      "avgScore"       -> List.sortBy .avgScore        model.data
                      _                -> List.sortBy .gender          model.data
      sortedData = if sortBy == "no" then model.data else
        case asc of
          True  -> sortedData1
          False -> List.reverse sortedData1
  in table []
      (List.concat
      [[ tr []
          [ th [] [ text "gender "
                  , Html.label [onClick <| ChangeVisualisation (Table "gender"          (if sortBy == "gender"          then not asc else True))]
                               [text (if sortBy == "gender"          then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "race "
                  , Html.label [onClick <| ChangeVisualisation (Table "race"            (if sortBy == "race"            then not asc else True))]
                               [text (if sortBy == "race"            then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "parentEducation "
                  , Html.label [onClick <| ChangeVisualisation (Table "parentEducation" (if sortBy == "parentEducation" then not asc else True))]
                               [text (if sortBy == "parentEducation" then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "lunch "
                  , Html.label [onClick <| ChangeVisualisation (Table "lunch"           (if sortBy == "lunch"           then not asc else True))]
                               [text (if sortBy == "lunch"           then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "testPreparation "
                  , Html.label [onClick <| ChangeVisualisation (Table "testPreparation" (if sortBy == "testPreparation" then not asc else True))]
                               [text (if sortBy == "testPreparation" then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "mathScore "
                  , Html.label [onClick <| ChangeVisualisation (Table "mathScore"       (if sortBy == "mathScore"       then not asc else True))]
                               [text (if sortBy == "mathScore"       then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "readingScore "
                  , Html.label [onClick <| ChangeVisualisation (Table "readingScore"    (if sortBy == "readingScore"    then not asc else True))]
                               [text (if sortBy == "readingScore"    then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "writingScore "
                  , Html.label [onClick <| ChangeVisualisation (Table "writingScore"    (if sortBy == "writingScore"    then not asc else True))]
                               [text (if sortBy == "writingScore"    then if asc then "▲" else "▼" else "⬍")]]
          , th [] [text "avgScore "
                  , Html.label [onClick <| ChangeVisualisation (Table "avgScore"        (if sortBy == "avgScore"        then not asc else True))]
                               [text (if sortBy == "avgScore"        then if asc then "▲" else "▼" else "⬍")]]
          ]
       ]
      , List.map (\data ->
          tr []
            [ td [] [text data.gender         ]
            , td [] [text data.race           ]
            , td [] [text data.parentEducation]
            , td [] [text data.lunch          ]
            , td [] [text data.testPreparation]
            , td [] [text <| String.fromInt data.mathScore   ]
            , td [] [text <| String.fromInt data.readingScore]
            , td [] [text <| String.fromInt data.writingScore]
            , td [] [text <| String.fromInt data.avgScore    ]
            ]
        ) sortedData
      ]
      )

-- viewHelper
viewHeader : Html Msg
viewHeader =
  div []
    [ Html.h1 [] [text "Students Performance In Exams"]
    , viewSelect "Visualisierung auswählen "
        [(ChangeVisualisation (Table "no" True)                        , "Tabelle"                                            )
        ,(ChangeVisualisation (QQ_Plot              .testPreparation 4), "QQ-Plot                          (Visualisierung 1)")
        ,(ChangeVisualisation (MultiDimPlot Nothing .parentEducation 2), "Multidimensionaler QQ-Plot       (Visualisierung 2)")
        ,(ChangeVisualisation ColorIcons                               , "Überblick mit gefärbten Symbolen (Visualisierung 3)")
        ]
    , hr [] []
    ]

viewSelect : String -> List (Msg, String) -> Html Msg
viewSelect name msgString =
    let tupleOfLists = List.unzip msgString
    in
    Html.div []
      [ Html.text name
      , Html.select []
           (List.map2
               (\msg string ->
                    Html.option [onClick msg]
                                [Html.text string]
               )
               (Tuple.first tupleOfLists) (Tuple.second tupleOfLists)
           )
      ]

-- Interface for Plots.Msg in html

viewInteractivePlot : Html Plots.Msg -> Html Msg
viewInteractivePlot html = Html.map (\plotMsg -> UpdatePlots plotMsg) html

-- Data preparation

calcQValues : Model -> (Data -> String) -> Int -> List XyData
calcQValues model selector dimIndex =
    let dimValues = list_get dataPossibleNominalValues dimIndex
    in  List.map (\dimValue ->
          let dataSelection = selectData selector dimValue model.data                             -- Select the data
          in  dataToQValuesXyData dataSelection.selected .avgScore .avgScore <| "with "++dimValue -- Calc Q-Values
        ) dimValues

dataToQValuesXyData : List Data -> (Data -> Int) -> (DataDescription -> String) -> String -> XyData
dataToQValuesXyData input compareDependency dependencyToString quantilesDescription =
    let sortedData = List.sortBy compareDependency input

        indexedData : List (Int,Data)
        indexedData = List.indexedMap Tuple.pair sortedData
    in XyData ("f-Value "  ++(dependencyToString dataDescription))
              ("quantiles "++(dependencyToString dataDescription)++" "++quantilesDescription)
         <| List.filterMap
              (\(i,data) -> Maybe.map3
                              Plots.pointConstructor
                              (Just "")
                              -- f_Value = i/n -- indexedMap builds i=0,...,n-1, but here I need i=0.5,...,n-0.5
                              (Just (((toFloat i)+0.5) / (toFloat (List.length sortedData))))
                              (Just <| toFloat <| compareDependency data)
              ) indexedData

    -- select
selectData : (Data -> String) -> String -> List Data -> {selected: List Data, unselected: List Data }
selectData dependencyFunction dependencyValue input =
    { selected   = List.filter
                     (\data -> (dependencyFunction data) == dependencyValue)
                     input
    , unselected = List.filter
                     (\data -> (dependencyFunction data) /= dependencyValue)
                     input
    }

-- Decode csv-String
checkInputFormat : {description: String, data: String} -> {description: String, data: String}
checkInputFormat input =
  if String.left 15 input.data == "<!DOCTYPE HTML>" then
    let -- Error handling for parsing
        deadEndsToString : List Parser.DeadEnd -> String -- sadly Parser.deadEndsToString always just prints "TODO: deadEndsToString"
        deadEndsToString deadEnds = "[ "++(List.foldl (++) " ]" <| List.map deadEndToString deadEnds)

        deadEndToString : Parser.DeadEnd -> String
        deadEndToString {row,col,problem} = "row = "++(String.fromInt row)++", col = "++(String.fromInt col)++", problem = "++(problemToString problem)

        problemToString : Parser.Problem -> String
        problemToString problem =
          case problem of
            Parser.Expecting string        -> "Expecting \""++string++"\""
            Parser.ExpectingInt            -> "ExpectingInt"
            Parser.ExpectingHex            -> "ExpectingHex"
            Parser.ExpectingOctal          -> "ExpectingOctal"
            Parser.ExpectingBinary         -> "ExpectingBinary"
            Parser.ExpectingFloat          -> "ExpectingFloat"
            Parser.ExpectingNumber         -> "ExpectingNumber"
            Parser.ExpectingVariable       -> "ExpectingVariable"
            Parser.ExpectingSymbol string  -> "ExpectingSymbol \""++string++"\""
            Parser.ExpectingKeyword string -> "ExpectingKeyword \""++string++"\""
            Parser.ExpectingEnd            -> "ExpectingEnd"
            Parser.UnexpectedChar          -> "UnexpectedChar"
            Parser.Problem string          -> "Problem \""++string++"\""
            Parser.BadRepeat               -> "BadRepeat"
        -- get the csv string from html document
        data = String.dropLeft 15 input.data -- the parser can't parse the <!DOCTYPE HTML>-part of Html
        htmlDataResult = Html.Parser.run data
        dataInCorrectFormat =
          case htmlDataResult of
            Ok htmlData ->
              let allTexts = findallTexts htmlData

                  findallTexts : List Html.Parser.Node -> List String
                  findallTexts nodes = List.concat <| List.map2 findallTextsHelper nodes <| List.repeat (List.length nodes) []
                  findallTextsHelper : Html.Parser.Node -> List String -> List String
                  findallTextsHelper outerNode texts =
                    case outerNode of
                      Html.Parser.Text string                     -> string::texts
                      Html.Parser.Element string attrs childNodes ->
                        List.concat
                          [ List.concat <| List.map2 findallTextsHelper childNodes <| List.repeat (List.length childNodes) []
                          , texts
                          ]
                      Html.Parser.Comment string                  -> []
              in -- The longest String should be the csv one
                 Maybe.withDefault "There is no text in this html-file." <| list_Last <| List.sortBy String.length allTexts
            Err errorList -> "Input was Html-file, but Error while Parsing it: "++(deadEndsToString errorList)++"\t\r\n"
                             ++"Html-file:\t\r\n"++data

    in {input | data = dataInCorrectFormat }
  else input

csvString_to_data : String -> List Data
csvString_to_data csvRaw =
    Csv.parse csvRaw
        |> Csv.Decode.decodeCsv decodeStudentsPerformanceInExams
        |> Result.toMaybe
        |> Maybe.withDefault []

decodeStudentsPerformanceInExams : Csv.Decode.Decoder ( Data -> a ) a
decodeStudentsPerformanceInExams =
  Csv.Decode.map (\gender race pE lunch test math read write -> Data gender race pE lunch test math read write ((math+read+write)//3))
    (Csv.Decode.field "gender" Ok
        |> Csv.Decode.andMap (Csv.Decode.field "race/ethnicity"              Ok)
        |> Csv.Decode.andMap (Csv.Decode.field "parental level of education" Ok)
        |> Csv.Decode.andMap (Csv.Decode.field "lunch"                       Ok)
        |> Csv.Decode.andMap (Csv.Decode.field "test preparation course"     Ok)
        |> Csv.Decode.andMap (Csv.Decode.field "math score"    (String.toInt >> Result.fromMaybe "error parsing math score"   ))
        |> Csv.Decode.andMap (Csv.Decode.field "reading score" (String.toInt >> Result.fromMaybe "error parsing reading score"))
        |> Csv.Decode.andMap (Csv.Decode.field "writing score" (String.toInt >> Result.fromMaybe "error parsing writing score"))
    )

-- working with custom types
visualisationKindToString : VisualisationKind -> String
visualisationKindToString vk =
  case vk of
    Table          _ _ -> "Table"
    QQ_Plot        _ _ -> "QQ_Plot"
    MultiDimPlot _ _ _ -> "MultiDimPlot"
    ColorIcons         -> "ColorIcons"